#define PIC16F676

#include "16f676.h"
#pragma config |= 0x318C // INT OSC; 318C with Watch Dog;

//#define Test_buzz 1

#pragma update_RP 0

typedef unsigned char byte;
typedef unsigned long int LONG;
typedef bit BOOL;

//--- SIGNAL DEFINITIONS -------------------------------------------------
byte Signal; // current signal active (0 = Not Active);
byte SignalOffset; // current signal sample position;
byte CurrSignal; // Counter of delay for current position;

#define TIMER_US 10000

//--- PORT DEFINITIONS ---------------------------------------------------
//               76543210
#define  TrisA 0b00001111                
//               ||||||||
//               |||||||CELL_1         
//               ||||||CELL_2         
//               |||||CELL_3        
//               ||||BUTTON
//               |||TEMP        
//               ||BUZZ
//               |---
//               ---

//               76543210
#define  TrisC 0b00000000
//               ||||||||
//               |||||||D4
//               ||||||D5
//               |||||D6 
//               ||||D7
//               |||E 
//               ||A0
//               |---
//               ---

// LCD connector;
// 1   2   3   4   5   6   7   8   9   10
// GND U+  A0  E Contr D5  D4  D7  D6  UBL+

#pragma  bit CELL_1   @ PORTA.0   //        - PIC-13
#pragma  bit CELL_2   @ PORTA.1   //        - PIC-12
#pragma  bit CELL_3   @ PORTA.2   //        - PIC-11
#pragma  bit TEMP     @ PORTA.4   //        - PIC-3 
#pragma  bit BUTTON   @ PORTA.3   //        - PIC-4 
#pragma  bit BUZZ     @ PORTA.5   //        - PIC-2 
#pragma  bit LCD_A0   @ PORTC.5   // LCD-4  - PIC-5 
#pragma  bit LCD_E    @ PORTC.4   // LCD-6  - PIC-6 
#pragma  bit LCD_D4   @ PORTC.0   // LCD-11 - PIC-10
#pragma  bit LCD_D5   @ PORTC.1   // LCD-12 - PIC-9
#pragma  bit LCD_D6   @ PORTC.2   // LCD-13 - PIC-8
#pragma  bit LCD_D7   @ PORTC.3   // LCD-14 - PIC-7

#define TEMP_mask 0b00010000

#define LCD_A0_mask 0b00100000
#define LCD_DD_mask 0b00001111

#define BUTTON_PRESSED 0
//------------------------------------------------------------------------

#define SIGNAL_VoltageLower 1
#define SIGNAL_VoltageUpper 2
#define SIGNAL_TempWarm 3
#define SIGNAL_TempHot 4

#include "int16CXX.H"

#define SET_RP0 \
#asm \
BSF   0x03,RP0 \
#endasm

#define CLR_RP0 \
#asm \
BCF   0x03,RP0 \
#endasm


byte GetSignalByOffset();       

#pragma origin 4
interrupt serverX( void)
{
/* W and STATUS are saved by the next macro. PCLATH is also
   saved if necessary. The code produced is CPU-dependent. */

int_save_registers    // W, STATUS (and PCLATH if required)
//char sv_FSR = FSR;  // save FSR if required
char sv_PCLATH = PCLATH; // save PCLATH

  // handle the interrupt (insert application code here)

   CLR_RP0;
   T1CON = 0b00000000; // Stop timer;

   if (Signal) { // Signal Active;
     if (CurrSignal) { // Still waiting;
      CurrSignal--;
     } else {
       CurrSignal = GetSignalByOffset();       
       if (!CurrSignal) { // if /x00, then loop to start;
         SignalOffset = 0; // return to 0;
         CurrSignal = GetSignalByOffset();       
       }
       if ((SignalOffset & 1) == 0) // first position -> BUZZ sounds;
         BUZZ = 1;
       else
         BUZZ = 0;
       SignalOffset++; // move forward;
       CurrSignal--;
     }
   }

   // set timer to wait TIMER_US=10 ms;
   TMR1L = ((65536-TIMER_US)) & 0xFF;
   TMR1H = ((65536-TIMER_US) >> 8) & 0xFF;
   T1CON = 0b00000001; // Resume counting;
   TMR1IF = 0;         

PCLATH = sv_PCLATH;    // restore PCLATH
//FSR = sv_FSR;       // restore FSR if saved
int_restore_registers // W, STATUS (and PCLATH if required)
}

byte Signals (byte i) {
skip(i);
#pragma return[] = 4, 7, 10, 15 	// 4, 4+3, 4+3+3, 4+3+3+5
#pragma return[] = 70,50,0		// SIGNAL_VoltageLower
#pragma return[] = 10,90,0		// SIGNAL_VoltageUpper
#pragma return[] = 15,10,15,90,0	// SIGNAL_TempWarm
#pragma return[] = 20,10,70,50,0	// SIGNAL_TempHot
}

byte GetSignalByOffset() {
  W = Signals(Signal-1);
  return Signals (W+SignalOffset);
}

void SetSignal (byte piSignal) {
  if (piSignal != Signal) {
    GIE = 0;
    Signal = piSignal; 
    BUZZ = 0;
    SignalOffset = 0;
    CurrSignal = 0;
    GIE = 1;
  }
}

#include "MATH16.H"

//--- PAUSE MAKING FUNCTIONS ---------------------------------------------
void Delay100us (byte aTs) // pause for aTs*100 us;
{ // 4+x*3 us = 4+3*32=100 us;
 do  {
     char i = 33;            /* 2 IS */
     do ; while ( --i > 0);  /* 26 * 3 - 1 = 77 IS */
 } while ( --aTs > 0);         /* 3 IS */
}

void Delay10ms (byte aTs) // pause for aTs*10 ms;
{
 for (; aTs > 0; aTs--) {
   Delay100us (100);
   clrwdt();
 }
}

/*
void DelayMs (unsigned char aTs) // pause for aTs*1 ms;
{
 for (; aTs > 0; aTs--) {
   clrwdt();
   Delay100us (10);
 }
}
*/

//--- LCD FUNCTIONS ------------------------------------------------------

#define LCD_DELAY_BASE_100_US 1

void lcdWrite (char c) {
  // high data byte;
  c = swap(c);
  PORTC = (c & LCD_DD_mask) | LCD_A0_mask; LCD_E = 1; nop(); LCD_E = 0;
  Delay100us (LCD_DELAY_BASE_100_US);
  // low data byte;
  c = swap(c);
  PORTC = (c & LCD_DD_mask) | LCD_A0_mask; LCD_E = 1; nop(); LCD_E = 0;
  Delay100us (LCD_DELAY_BASE_100_US);
}

void lcdCommand (char c) {
  // high data byte;
  c = swap(c);
  PORTC = (c & LCD_DD_mask); LCD_E = 1; nop(); LCD_E = 0;
  Delay100us (LCD_DELAY_BASE_100_US);
  // low data byte;
  c = swap(c);
  PORTC = (c & LCD_DD_mask); LCD_E = 1; nop(); LCD_E = 0;
  Delay100us (LCD_DELAY_BASE_100_US);
}

void lcdCommand8 (char c) {
  PORTC = c; LCD_E = 1; nop(); LCD_E = 0;
  Delay100us (LCD_DELAY_BASE_100_US);
}

#define LCD_CMD_POS1		( 0x80 )
#define LCD_CMD_POS2		(0x80 | 0x40)

#define LCD_CURSOR_ON		lcdCommand (0xC | 0x2);
#define LCD_CURSOR_OFF		lcdCommand (0xC);
#define LCD_CURSOR_BLINK_ON	lcdCommand (0xC | 0x2 | 0x1);
#define LCD_CLEAR		{ lcdCommand (0x1); Delay100us (15+LCD_DELAY_BASE_100_US); }
#define LCD_CURSOR_HOME		lcdCommand (0x2);
#define LCD_SHIFT_CURSOR_LEFT	lcdCommand (0x10);
#define LCD_SHIFT_CURSOR_RIGHT	lcdCommand (0x10 | 0x4);
#define LCD_SHIFT_DISPLAY_LEFT	lcdCommand (0x10 | 0x8);
#define LCD_SHIFT_DISPLAY_RIGHT	lcdCommand (0x10 | 0x8 | 0x4);
#define LCD_CURSOR_LINE_1(x)	lcdCommand (LCD_CMD_POS1 | x);
#define LCD_CURSOR_LINE_2(x)	lcdCommand (LCD_CMD_POS2 | x);

void lcdProgram (const char * p) {
  byte i;
  BOOL lbCommand = 1;
  char c;
  for (i = 0; ; i++) {
    char c = p[i];
    if (c == 0)
      return;
    if (lbCommand) {
      lcdCommand (c);
      if (c & LCD_CMD_POS1) // output character after cursor placement command;
        lbCommand = 0;
    } else {
      lcdWrite (c);
      lbCommand = 1;
    }
  }
}

void lcdWriteln (const char * p) {
  byte i;
  for (i = 0; p[i] != 0; i++) {
      lcdWrite(p[i]);
  }
}

const char LCD_Data_Init[] = {
  0x28, //  
  0x0C, //  
  0x6,  //     
  0x0F, //     
//  0x80, //  
//  ' ',
  0x01,  // ,  
  0
};

void LCD_Init() { // 16s2d
  PORTC = 0;
  Delay10ms (5);

// 30h  
  lcdCommand8 (3);
  lcdCommand8 (3);
  lcdCommand8 (3);

// 20h  
  lcdCommand8 (2);

//4bit
  lcdProgram (LCD_Data_Init);
  Delay100us (15);
}

void WriteNum (LONG n, BOOL Point) {
  n = n+5;
  byte i = '0';
  while (n >= 100) {
    n -= 100;
    i++;
  }
  lcdWrite (i);
  if (Point) {
    lcdWrite ('.');  
  }

  i = '0';
  while (n >= 10) {
    n -= 10;
    i++;
  }
  lcdWrite (i);
}

/*
void Wait () {
 Delay10ms(50);
 Delay10ms(50);
}
*/

//--- ANALOG SETUP DATA --------------------------------------------------
#define ADChannel_Cell1 	0
#define ADChannel_Cell2		1
#define ADChannel_Cell3		2
#define ADChannel_Temp 		3

LONG Channels[4];
LONG Cell[3];
LONG Temp;

struct Limit {
 LONG Lower;
 LONG Upper; 
};

#define DATA_LIMIT_ADDR 1

byte limit;
struct Limit Voltage;

#define TempExtCount 6

struct TempExtData {
  LONG T_10;
  LONG V_16;
  byte K;
};

const struct TempExtData TempExt[TempExtCount] =
{{200, 	3389, 	99},
{360, 	3803, 	91},
{520, 	4252, 	85},
{680, 	4734, 	80},
{840, 	5246, 	76},
{1000, 	5787, 	76}};

#define VoltageCount 8

#define DEFAULT_LIMIT 5

const struct Limit Voltages[VoltageCount] =
{{250, 270}, // limit = 0, beep: -
{260, 280},  // limit = 1, beep: - -
{270, 290},  // limit = 2, beep: - - -
{280, 300},  // limit = 3, beep: - - - --
{290, 310},  // limit = 4, beep: - - - -- -
{300, 320},  // limit = 5, beep: - - - -- - -
{310, 330},  // limit = 6, beep: - - - -- - - -
{320, 340}}; // limit = 7, beep: - - - -- - - - --

#define Temp_Lower 500
#define Temp_Upper 600

#define NUM_VOLTAGE_SAMPLES     8  // 2, 4, 8, 16

LONG ReadADC(byte piADChannel) {
  if (piADChannel == ADChannel_Temp) { // enable the termistor;
    SET_RP0;
    ANSEL = (1 << ADChannel_Cell1) | (1 << ADChannel_Cell2) | (1 << ADChannel_Cell3) | (1 << ADChannel_Temp);
    TRISA = TrisA | TEMP_mask;
  }
  CLR_RP0;
  byte a = (piADChannel << 2);
  ADCON0 = (ADCON0 & 0b111.000.11) | a;
  Delay100us(10);
  GO = 1;
  while (GO) { };

  // disable the termistor by powering TEMP pin;
  SET_RP0;
  ANSEL = (1 << ADChannel_Cell1) | (1 << ADChannel_Cell2) | (1 << ADChannel_Cell3);
  TRISA = TrisA;
  CLR_RP0;
  TEMP = 1;

  long unsigned int u;
  u.high8 = ADRESH;
  SET_RP0;
  u.low8 = ADRESL;
  CLR_RP0;
  return u;
}

void ReadChannels() {
  Channels[0] = 0;
  Channels[1] = 0;
  Channels[2] = 0;
  Channels[3] = 0;
  signed int i = NUM_VOLTAGE_SAMPLES;
  while (i > 0) {
    Channels[0] += ReadADC(ADChannel_Cell1);
    Channels[1] += ReadADC(ADChannel_Cell2);
    Channels[2] += ReadADC(ADChannel_Cell3);
    Channels[3] += ReadADC(ADChannel_Temp);
    i--;
  }

  Channels[0] /= NUM_VOLTAGE_SAMPLES; // 1:1
  Channels[1] /= NUM_VOLTAGE_SAMPLES; // 1:2
  Channels[2] /= NUM_VOLTAGE_SAMPLES; // 1:3
  Channels[3] /= NUM_VOLTAGE_SAMPLES;
  // cell_1 = Channels(0) * 100 * 5 / 1023;
  // cell_2 = (Channels(1)*2-Channels(0)) * 100 * 5 / 1023;
  // cell_3 = (Channels(2)*3-Channels(1)*2) * 100 * 5 / 1023;
  
  // cell_1 = x = Channels[0]; x*(0.5-3/256) = x*(0,48828125) ~= 100 * 5 / 1023 = 0,488758553; 
  Cell[0] = Channels[0]/2; 	// x/2
  LONG p1 = Channels[0]*3; 	// x*3
  Cell[0] -= p1.high8;

  // cell_2 = x = Channels[1]; x*(1-6/256) = x*(0,9765625) ~= 100 * 5 / 1023 * 2= 0,977517107;
  Cell[1] = Channels[1]; 	// x
  p1 = Channels[1]*6; 		// x*6
  Cell[1] -= p1.high8;        	// x-6*x/256

  // cell_3 = x = Channels[2]; x*(1,5-8,5/256)=x*(1,466796875) ~= 100*5/1023*3 = 1,46627566;
  Cell[2] = Channels[2]/2; 	// x/2
  p1 = Channels[2]*8; 		// x*8
  p1+= Cell[2];                 // x * 8,5;
  Cell[2] = Channels[2]/2;      // x/2
  Cell[2] += Channels[2];       // 1,5*x
  Cell[2] -= p1.high8; 		// 1,5 * x- 8,5 * x/256

  if (Channels[3] < 5) {
    Temp = 0;
  } else {
    p1 = Channels[3]*16; // V*16;
    for (i = TempExtCount-1; i >= 0; i--) {
      LONG Vb_16 = TempExt[i].V_16;
      if ((p1 > Vb_16) || (i == 0)) {
        p1 -= Vb_16;
        byte K = TempExt[i].K;
        p1 *= K;
        p1 += 128;
        Temp = p1.high8;
        p1 = TempExt[i].T_10;
        Temp += p1;
        break;
      }
    }  
  }

  Cell[2] -= Cell[1];
  Cell[1] -= Cell[0];
}

// Find if channels Cell1, Cell2, Cell3 and Temp are available;
void ReadAndOutputAvailableChannels() {
  ReadChannels();
  LCD_CURSOR_LINE_1 (0);  WriteNum (Cell[0], 1);
  LCD_CURSOR_LINE_1 (4);  WriteNum (Cell[1], 1);
  LCD_CURSOR_LINE_1 (8);  WriteNum (Cell[2], 1);
  LCD_CURSOR_LINE_1 (13); WriteNum (Temp, 0);
}

void CheckAndSignal() {
  if (Temp >= Temp_Upper) {
    SetSignal (SIGNAL_TempHot);
    return;
  }
  
  byte liSignal = 0;
  byte i;
  for (i=0; i<3; i++) {
    LONG c = Cell[i];
    if (c <= Voltage.Lower) {
      SetSignal (SIGNAL_VoltageLower);
      return;
    }
    if (c <= Voltage.Upper) {
      liSignal = SIGNAL_VoltageUpper;
    }
  }
  if (Temp >= Temp_Lower) {
    liSignal = SIGNAL_TempWarm;
  } 
  SetSignal (liSignal);
}

/********************************* EEPROM Read and Write *******************************************/

#define DATA_VALID 200
#define DATA_VALID_ADDR 0

// Write data to the EEPROM
// Beware - it takes about 4 mS for the write to complete
void EEPROM_Write(char addr, char data) {
	EEIF = 0;					// clear done flag
        SET_RP0;
	EEADR = addr;					// set up address
#if defined PIC16F676
	EEDAT = data;					// set up data to write
#else
	EEDATA = data;					// set up data to write
#endif
	WREN = 1;					// enable writing
	EECON2 = 0x55;					// required sequence
	EECON2 = 0xAA;
	WR = 1;						// do the write
	WREN = 0;					// disable other writes
        CLR_RP0;
	while (EEIF == 0) {};				// wait for write to complete - about 4mS
	EEIF = 0;					// clear done flag
}

char EEPROM_Read(char addr) {
	char dat;
        SET_RP0;
	EEADR = addr;
	RD = 1;						// do the read
#if defined PIC16F676
	dat = EEDAT;
#else
	dat = EEDATA;
#endif
        CLR_RP0;
	return dat;					// result immediately available
}

#ifdef Test_buzz
byte iSignal;
#endif

void SetLimit() {
 Voltage.Lower = Voltages[limit].Lower;
 Voltage.Upper = Voltages[limit].Upper;
 EEPROM_Write (DATA_LIMIT_ADDR, limit);

 LCD_CURSOR_LINE_2 (0);  WriteNum (Voltage.Lower, 1);
 LCD_CURSOR_LINE_2 (4);  WriteNum (Voltage.Upper, 1);
 LCD_CURSOR_LINE_2 (10); WriteNum (Temp_Lower, 0);
 LCD_CURSOR_LINE_2 (13); WriteNum (Temp_Upper, 0);

 BUZZ = 0;
 if (Signal)
   Delay10ms (50);

 int i = 1;
 while (i <= limit+1) {
   BUZZ = 1;
   Delay10ms (20);
   if ((i & 3) == 0)
     Delay10ms (20);
   BUZZ = 0;
   Delay10ms (20);
   i++;
 }
 if (Signal)
   Delay10ms (30);

#ifdef Test_buzz
 if (iSignal == 0)
   SetSignal (0);
 else 
 if (iSignal == 1)
   SetSignal (SIGNAL_VoltageLower);
 else 
 if (iSignal == 2)
   SetSignal (SIGNAL_VoltageUpper);
 else 
 if (iSignal == 3)
   SetSignal (SIGNAL_TempWarm);
 else 
   SetSignal (SIGNAL_TempHot);
 LCD_CURSOR_LINE_1 (12);  lcdWrite ('0'+Signal);

 iSignal++;
 if (iSignal > 4) 
   iSignal = 0;
#endif
}

void Press() {
 limit++;
 if (limit >= VoltageCount) {
   limit = 0;
 } 
 SetLimit();
}

#define LCD_CH_ALARM		0xED
#define LCD_CH_TEMPERATURE	0xEF
#define LCD_CH_ACCURE_DATA      0xED

const char LCD_Data_Fill[] = {
  LCD_CMD_POS1 | 11, 'V',
  LCD_CMD_POS1 | 15, LCD_CH_TEMPERATURE,

  LCD_CMD_POS2 | 3,  '/',
  LCD_CMD_POS2 | 7,  'V',
  LCD_CMD_POS2 | 12, '/',
  LCD_CMD_POS2 | 15, LCD_CH_TEMPERATURE,
  0
};

char GetOsccal();							// Defined at end of file

void main() {
  W = GetOsccal();						// Set built-in oscillator calibration
  SET_RP0;
  OSCCAL = W;
  
  TRISA = TrisA;
  TRISC = TrisC;
  OPTION = 0b11001110; // no pull-up, int/b0 on 0->1; Prescalder 64 -> ~1 sec. to WDT;
  CLR_RP0;
  PORTA = 0; 
  TEMP = 1; // disable termistor with connecting its both pins to the 1. 5 ma current will flow through the TEMP pin.
  PORTC = 0; 

  CMCON = 0b.0000.0111; // comparator is turned off;

  SET_RP0;
  ANSEL = (1 << ADChannel_Cell1) | (1 << ADChannel_Cell2) | (1 << ADChannel_Cell3);

  ADCON1 = 0b0.101.0000;      //  Select the Clock as Fosc/16

  ADIE = 0;

  CLR_RP0;
  ADCON0 = 0b10.0.000.01;     //  Turn on the ADC
                              //   Bit 7 - Right Justified Sample
                              //   Bit 6 - Dont Use VRef
                              //   Bit 4:2 - Channel 0
                              //   Bit 1 - Do not Start
                              //   Bit 0 - Turn on ADC

  Signal = 0; // Not Active;
#ifdef Test_buzz
  iSignal = 0;
#endif

  LCD_Init();          
//  LCD_CURSOR_LINE_1 (0);
//  lcdWriteln ("3 LiPo V/t");
  lcdWriteln ("3 LiPo V/t Alarm");
//  LCD_CURSOR_LINE_2 (0);
//  lcdWriteln ("(c) Prikupets A.");
  Delay10ms (100);
  LCD_CLEAR;

  lcdProgram (LCD_Data_Fill);
  ReadAndOutputAvailableChannels();

  if (EEPROM_Read(DATA_VALID_ADDR) != DATA_VALID) {
    limit = DEFAULT_LIMIT;
    EEPROM_Write (DATA_VALID_ADDR, DATA_VALID);
  } else {
    limit = EEPROM_Read(DATA_LIMIT_ADDR);
  }
  SetLimit();

  // Init Timer1;
  TMR1L = ((65536-TIMER_US)) & 0xFF;
  TMR1H = ((65536-TIMER_US) >> 8) & 0xFF;
  T1CON = 0b00000001; // Resume counting;
  TMR1IF = 0;           

  SET_RP0;
  PIE1 = 1; 

  PEIE = 1;
  GIE = 1;

  CLR_RP0;
  BOOL WaitButton = (BUTTON==BUTTON_PRESSED);
  byte iInertial = 0;
  BOOL lbValidation = 0;
  while (1) {
    clrwdt(); 
    ReadAndOutputAvailableChannels();

#define VALIDATION_COUNT 25

    lcdCommand (LCD_CMD_POS2 | 9);
    if (iInertial == VALIDATION_COUNT) { // Validation each ~ 1s;
      CheckAndSignal();
      lcdWrite (LCD_CH_ACCURE_DATA);
      iInertial = 0;
    } else 
    if (iInertial > VALIDATION_COUNT/4) {
      lcdWrite (' ');
    }
    iInertial++;

    if (WaitButton) {
      lcdCommand (LCD_CMD_POS2 | 7);
      LCD_CURSOR_BLINK_ON;
    } else {
      LCD_CURSOR_OFF;
    }

    if ((BUTTON==BUTTON_PRESSED) && !WaitButton) {
      Delay10ms(2);
      if (BUTTON==BUTTON_PRESSED) {
        Press();
        WaitButton = 1;
      }
    } else
    if ((BUTTON != BUTTON_PRESSED) && WaitButton) {
      Delay10ms(2);
      if (BUTTON != BUTTON_PRESSED) {
        WaitButton = 0;
      }
    }
  };
}

#pragma origin 0x3FF
// This is a dummy - actual procedure is preset in chip and will override
// A change of 1 causes about 1% change in frequency
char GetOsccal() {return 0x0;}

/*
// Initialize EEPROM data
#define EEPROM_START 0x2100
#pragma cdata[EEPROM_START]
#pragma cdata[] = MAGIC, 0x00	
*/
